import pandas as pd
import sqlite3
from eric_chen_forward import util
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer
from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report
from sklearn.metrics import accuracy_score, confusion_matrix
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import SGDClassifier
from sklearn.linear_model import LogisticRegression
import pickle

class Classifier:

    def __init__(self) -> None:
        self.eda_util = util.EDA(alpha=0.2)
        self.categories = []

    def train(self, labels_file, passages_file, csv_file=None):
        if csv_file:
            df = pd.read_csv(csv_file)
            labels = df['label'].to_list()
            passages = df['passage'].to_list()
        else:
            with open(labels_file) as f:
                labels = f.read().splitlines()
            with open(passages_file) as f:
                passages = f.read().splitlines()

        eda_passages = self.eda_util.run_data_augmentation(passages)

        passages.extend(eda_passages)
        labels.extend(labels)
        assert(len(passages) == len(labels))

        self.categories = list(set(labels))

        eda_df = pd.DataFrame()
        eda_df['label'] = labels
        eda_df['passage'] = passages
        eda_df['cleaned_text'] = eda_df['passage'].apply(lambda x: util.clean_document(x))
        eda_df['num_years'] = eda_df['passage'].apply(lambda x: util.num_years(x))

        X = eda_df[['cleaned_text', 'num_years']]
        y = eda_df['label']
        column_transformer = ColumnTransformer([
            ('tfidf', TfidfVectorizer(), 'cleaned_text')
        ])
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.33, random_state=0, stratify=y)

        sgd = Pipeline([('preprocess', column_transformer),
                ('clf', SGDClassifier(loss='hinge', penalty='l2', alpha=1e-3, random_state=0, tol=None))])
        sgd.fit(X_train, y_train)
        y_pred = sgd.predict(X_test)
        print('SVM with SGD')
        print(f'accuracy: {accuracy_score(y_pred, y_test)}')
        print(classification_report(y_test, y_pred))
        print('-' * 55)

        log_reg = Pipeline([('preprocess', column_transformer),
                ('clf', LogisticRegression(C=1e5))])

        log_reg.fit(X_train, y_train)
        y_pred = log_reg.predict(X_test)
        print('Logistic Regression')
        print(f'accuracy: {accuracy_score(y_pred, y_test)}')
        print(classification_report(y_test, y_pred))
        print('-' * 55)

        choice = input('Which model to save (SVM or LR): ').strip().upper()
        while choice != 'SVM' and choice != 'LR':
            print('Invalid selection.')
            choice = input('Which model to save (SVM or LR): ').upper()

        model_name = input('Enter model name: ')
        with open(f'{model_name}.pkl', 'wb') as f:
            if choice == 'SVM':
                pickle.dump(sgd, f)
            elif choice == 'LR':
                pickle.dump(log_reg, f)
            print(f'Saving {choice} model at {model_name}.pkl')

    def get_categories(self):
        return self.categories



import sqlite3

class Database:

    def __init__(self) -> None:
        self.db_name = "sqlite3_database.db"

    def insert_initial_classifier_data(labels_file, passages_file):
        query = """
             CREATE TABLE "dataset" (
                "id"	INTEGER NOT NULL,
                "labels"	TEXT NOT NULL,
                "passages"	TEXT NOT NULL,
                PRIMARY KEY("id" AUTOINCREMENT)
            );
        
        """



from eric_chen_forward import util
import streamlit as st
import pickle
import requests
from bs4 import BeautifulSoup
import pandas as pd
from trafilatura import fetch_url, extract
from trafilatura.settings import use_config
from transformers import AutoTokenizer, BartForConditionalGeneration

def Demo(max_summary_length=100):
    with open('model.pkl', 'rb') as f:
        model = pickle.load(f)

    def extract_paragraphs(url, use_trafilatura=False):
        if use_trafilatura:
            config = use_config()
            config.set("DEFAULT", "EXTRACTION_TIMEOUT", "0")
            downloaded = fetch_url(url)
            result = extract(downloaded, config=config)
            return result
        else:
            data = requests.get(url).text
            soup = BeautifulSoup(data, 'lxml')
            h = soup.find('header')
            if h:
                h.decompose()
            f = soup.find('footer')
            if f:
                f.decompose()
            if soup.body is None:
                return None
            
            paragraphs = []
            for p in soup.find_all('p'):
                if len(p) > 0 and 'block' not in p:
                    text = p.get_text(strip=True, separator='\n')
                    if len(paragraphs) > 0 and len(paragraphs[-1].split()) < 100:
                        paragraphs[-1] += text
                    else:
                        paragraphs.append(text)
            return paragraphs



    title = st.text_input(label="Enter url", placeholder='url')

    tab1, tab2, tab3, tab4 = st.tabs(['BeautifulSoup', 'Trafilatura', 'BeautifulSoup + Summarization', 'Trafilatura + Summarization'])

    bart_model = BartForConditionalGeneration.from_pretrained("facebook/bart-large-cnn")
    tokenizer = AutoTokenizer.from_pretrained("facebook/bart-large-cnn")

    if title:
        paragraphs1 = extract_paragraphs(title, use_trafilatura=False)
        paragraphs2 = extract_paragraphs(title, use_trafilatura=True)
        with tab1:
            for paragraph in paragraphs1:
                cleaned_text = util.clean_document(paragraph)
                if len(cleaned_text) == 0:
                    continue
                temp = pd.DataFrame()
                temp['cleaned_text'] = [cleaned_text]
                temp['num_years'] = [util.num_years(paragraph)]
                pred = model.predict(temp[['cleaned_text', 'num_years']])
                
                st.write("Predicted Category: " + pred[0])
                st.write(paragraph)
                st.markdown('''---''')
        with tab2:
            cleaned_text = util.clean_document(paragraphs2)
            temp = pd.DataFrame()
            temp['cleaned_text'] = [cleaned_text]
            temp['num_years'] = [util.num_years(paragraphs2)]
            pred = model.predict(temp[['cleaned_text', 'num_years']])
            
            st.write("Predicted Category: " + pred[0])
            st.write(paragraphs2)
            st.markdown('''---''')
        with tab3:
            for paragraph in paragraphs1:
                inputs = tokenizer([paragraph], max_length=1024, return_tensors="pt")
                summary_ids = bart_model.generate(inputs["input_ids"], num_beams=2, min_length=0, max_length=max_summary_length)
                summary = tokenizer.batch_decode(summary_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False, truncation=True)[0]
                cleaned_text = util.clean_document(summary)
                if len(cleaned_text) == 0:
                    continue
                temp = pd.DataFrame()
                temp['cleaned_text'] = [cleaned_text]
                temp['num_years'] = [util.num_years(summary)]
                pred = model.predict(temp[['cleaned_text', 'num_years']])
                
                st.write("Predicted Category: " + pred[0])
                st.write(summary)
                st.markdown('''---''')
        with tab4:
            inputs = tokenizer([paragraphs2], max_length=1024, return_tensors="pt")
            summary_ids = bart_model.generate(inputs["input_ids"], num_beams=2, min_length=0, max_length=max_summary_length)
            summary = tokenizer.batch_decode(summary_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False, truncation=True)[0]
            cleaned_text = util.clean_document(summary)
            temp = pd.DataFrame()
            temp['cleaned_text'] = [cleaned_text]
            temp['num_years'] = [util.num_years(summary)]
            pred = model.predict(temp[['cleaned_text', 'num_years']])
            
            st.write("Predicted Category: " + pred[0])
            st.write(summary)
            st.markdown('''---''')


import re
from nltk.corpus import stopwords
from nltk.tokenize import word_tokenize
from nltk.stem import WordNetLemmatizer
from nltk.corpus import wordnet
import random
import os
import spacy
from spacy import displacy

def clean_document(strs):
    strs = strs.replace('\\n', '').replace('\\r', '') # newline
    strs = re.sub('[^a-zA-Z]', ' ', strs) # punctuation
    strs = ' '.join(re.split('\s+', strs)) # duplicate whitespace
    strs = strs.lower().strip() # lowercase
    
    stop_words = set(stopwords.words('english'))
    word_tokens = word_tokenize(strs)
    lemmatizer = WordNetLemmatizer()
    filtered = [lemmatizer.lemmatize(w) for w in word_tokens if w not in stop_words]
    
    return ' '.join(filtered)


def num_folders(rootdir):
    s = set()
    for subdir, dirs, files in os.walk(rootdir):
        for file in files:
            path = os.path.join(subdir, file).split('/')
            s.add(path[1])

    return len(s)


def num_years(sentence):
    strs = sentence.split(' ')
    count = 0
    for s in strs:
        match = re.match(r'.*([1-2][0-9]{3})', s)
        if match is not None:
            count += 1
    
    return count

def replace_years(sentence):
    return ' '.join([re.sub(r'.*([1-2][0-9]{3})', 'YEAR', s) for s in sentence.split()])


def ner_count(passage):
    ner = spacy.load("en_core_web_sm")
    names = set()
    dates = []

    tags = ner(passage)
    for entity in tags.ents:
        text = entity.text
        label = entity.label_

        if label == 'PERSON':
            names.add(text)
        elif label == 'DATE':
            dates.append(text)
    return len(names), len(dates)


# Easy Data Augmentation implementation
class EDA:
    
    def __init__(self, alpha) -> None:
        self.alpha = alpha
        self.p = alpha
        self.stop_words = stopwords.words('english')

    def run_data_augmentation(self, passages):
        def find_synonyms(word):
            synonyms = []
            for syn in wordnet.synsets(word):
                for l in syn.lemmas():
                    synonyms.append(l.name())
            
            return list(set(synonyms))

        def synonym_replacement(sentence):
            words = sentence.split()
            n = int(self.alpha * len(words))
            idxs = [i for i in range(len(words)) if words[i] not in self.stop_words]
            idxs = random.sample(idxs, n)
            for i in idxs:
                synonyms = find_synonyms(words[i])
                if len(synonyms) > 0:
                    words[i] = random.choice(synonyms)
            
            return " ".join(words)

        def random_insertion(sentence):
            words = sentence.split()
            n = int(self.alpha * len(words))
            idxs = [i for i in range(len(words)) if words[i] not in self.stop_words]
            for i in range(n):
                idx = random.choice(idxs)
                word = words[idx]
                synonyms = find_synonyms(words[i])
                synonym = random.choice(synonyms) if len(synonyms) > 0 else None
                if synonym is not None:
                    idx = random.choice(range(len(words)))
                    words.insert(idx, synonym)
            
            return " ".join(words)

        def random_swap(sentence):
            words = sentence.split()
            n = int(self.alpha * len(words))
            for i in range(n):
                a, b = random.sample(range(len(words)), 2)
                words[a], words[b] = words[b], words[a]
            
            return " ".join(words)

        def random_deletion(sentence):
            words = sentence.split()
            words[:] = [word for word in words if random.uniform(0, 1) > self.p]
            return " ".join(words)
        
        
        eda_passages = []
        for passage in passages:
            sentences = passage.split('.')
            new_passage = ""
            for sentence in sentences:
                f = random.randint(0, 3)
                if f == 0:
                    sentence = synonym_replacement(sentence)
                if f == 1:
                    sentence = random_insertion(sentence)
                if f == 2:
                    sentence = random_swap(sentence)
                if f == 3:
                    sentence = random_deletion(sentence)
                new_passage += sentence + ". "
            eda_passages.append(new_passage)
        return eda_passages
    